from jupyter_protocol.messages import *

def test_create_simple():
    msg = Message.from_type('kernel_info_request', {})
    assert isinstance(msg.content, dict)
    assert isinstance(msg.metadata, dict)
    assert isinstance(msg.header, dict)
    assert isinstance(msg.parent_header, dict)
    assert isinstance(msg.header['msg_id'], str)
    assert msg.header['msg_type'] == 'kernel_info_request'



import logging
import os
import pprint
import zmq
from .session import Session

def app_or_module_logger(module_name):
    """
    If a global Application is instantiated, grab its logger.
    Otherwise, get a logger for the module name.
    """
    from traitlets.config import Application
    if Application.initialized():
        return Application.instance().log
    else:
        _logger = logging.getLogger(module_name)
        # Add a NullHandler to silence warnings about not being
        # initialized, per best practice for libraries.
        _logger.addHandler(logging.NullHandler())
        return _logger


# singleton dummy tracker, which will always report as done
DONE = zmq.MessageTracker()


class MessagingBase:
    # Whether to check PID to protect against calls after fork.
    # This check can be disabled if fork-safety is handled elsewhere.
    check_pid = True

    # Threshold (in bytes) beyond which a buffer should be sent without copying.
    copy_threshold = 2 ** 16

    debug = False  # Turn on for debugging output

    def __init__(self, connection_info):
        self.connection_info = connection_info
        #self.manager = manager
        #self.using_heartbeat = use_heartbeat and (manager is not None)
        self.log = app_or_module_logger(__name__)
        self.pid = os.getpid()
        self.context = zmq.Context.instance()
        self.session = Session(key=connection_info['key'].encode('ascii'),
                               signature_scheme=connection_info[
                                   'signature_scheme'])
        self.sockets = {}

    def _make_url(self, channel):
        """Make a ZeroMQ URL for a given channel."""
        transport = self.connection_info['transport']
        ip = self.connection_info['ip']
        port = self.connection_info['%s_port' % channel]

        if transport == 'tcp':
            return "tcp://%s:%i" % (ip, port)
        else:
            return "%s://%s-%s" % (transport, ip, port)

    def _create_bound_socket(self, channel):
        url = self._make_url(channel)
        socket_type = self.socket_types[channel]
        self.log.debug("Binding {} socket to: {}".format(channel, url))
        sock = self.sockets[channel] = self.context.socket(socket_type)
        # set linger to 1s to prevent hangs at exit
        sock.linger = 1000
        if url.startswith('tcp://') and url.endswith(':0'):
            port = sock.bind_to_random_port(url[:-2])
            self.connection_info['%s_port' % channel] = port
        else:
            sock.bind(url)
        return sock

    def recv(self, channel, mode=zmq.NOBLOCK, content=True, copy=True):
        """Receive and unpack a message.

        Parameters
        ----------
        socket : ZMQStream or Socket
            The socket or stream to use in receiving.

        Returns
        -------
        [idents], msg
            [idents] is a list of idents and msg is a nested message dict of
            same format as self.msg returns.
        """
        try:
            raw_msg = self.sockets[channel].recv_multipart(mode, copy=copy)
        except zmq.ZMQError as e:
            if e.errno == zmq.EAGAIN:
                # We can convert EAGAIN to None as we know in this case
                # recv_multipart won't return None.
                return None
            else:
                raise

        print('recvd raw', raw_msg)
        return self.session.deserialize(raw_msg, content=content, copy=copy)

    def send(self, channel, msg, *, track=False):
        """Send a message via stream or socket.

        The message format used by this function internally is as follows:

        [ident1,ident2,...,DELIM,HMAC,p_header,p_parent,p_content,
         buffer1,buffer2,...]

        The serialize/deserialize methods convert the nested message dict into this
        format.

        Parameters
        ----------

        channel : str
            The name of a socket to send the data on (e.g. 'shell')
        msg : Message
            A message object to send
        track : bool
            Whether to track.  Only for use with Sockets, because ZMQStream
            objects cannot track messages.
        """
        if self.check_pid and not os.getpid() == self.pid:
            app_or_module_logger(__name__).warning(
                "WARNING: attempted to send message from fork\n%s", msg
            )
            return

        socket = self.sockets[channel]

        # Check buffers are acceptable
        for idx, buf in enumerate(msg.buffers):
            if isinstance(buf, memoryview):
                view = buf
            else:
                try:
                    # check to see if buf supports the buffer protocol.
                    view = memoryview(buf)
                except TypeError:
                    raise TypeError("Buffer objects must support the buffer protocol.")
            # memoryview.contiguous is new in 3.3,
            # just skip the check on Python 2
            if hasattr(view, 'contiguous') and not view.contiguous:
                # zmq requires memoryviews to be contiguous
                raise ValueError("Buffer %i (%r) is not contiguous" % (idx, buf))

        to_send = self.session.serialize(msg)
        #print(to_send)
        longest = max([ len(s) for s in to_send ])
        copy = (longest < self.copy_threshold)

        if msg.buffers and track and not copy:
            # only really track when we are doing zero-copy buffers
            tracker = socket.send_multipart(to_send, copy=False, track=True)
        else:
            # use dummy tracker, which will be done immediately
            tracker = DONE
            socket.send_multipart(to_send, copy=copy)

        if self.debug:
            pprint.pprint(msg)
            pprint.pprint(to_send)

        msg.tracker = tracker

    def close(self):
        """Close all sockets"""
        for sock in self.sockets.values():
            sock.close()

class ClientMessaging(MessagingBase):
    socket_types = {
        'hb' : zmq.REQ,
        'shell' : zmq.DEALER,
        'iopub' : zmq.SUB,
        'stdin' : zmq.DEALER,
        'control': zmq.DEALER,
        'nanny_events': zmq.SUB,
        'nanny_control': zmq.DEALER,
    }

    def __init__(self, connection_info, use_heartbeat=True):
        super().__init__(connection_info)

        identity = self.session.session_id.encode('utf-8')
        self.iopub_socket = self._create_connected_socket('iopub', identity)
        self.iopub_socket.setsockopt(zmq.SUBSCRIBE, b'')
        self.shell_socket = self._create_connected_socket('shell', identity)
        self.stdin_socket = self._create_connected_socket('stdin', identity)
        self.control_socket = self._create_connected_socket('control', identity)

        if 'nanny_control_port' in connection_info:
            self.nanny_control_socket = self._create_connected_socket('nanny_control', identity)
            self.nanny_events_socket = self._create_connected_socket('nanny_events', identity)

    def _create_connected_socket(self, channel, identity=None):
        """Create a zmq Socket and connect it to the kernel."""
        url = self._make_url(channel)
        socket_type = self.socket_types[channel]
        self.log.debug("Connecting to: %s" % url)
        sock = self.sockets[channel] = self.context.socket(socket_type)
        # set linger to 1s to prevent hangs at exit
        sock.linger = 1000
        if identity:
            sock.identity = identity
        sock.connect(url)
        return sock


class KernelMessaging(MessagingBase):
    socket_types = {
        'hb': zmq.REP,
        'shell': zmq.ROUTER,
        'iopub': zmq.PUB,
        'stdin': zmq.ROUTER,
        'control': zmq.ROUTER,
    }

    def __init__(self, connection_info):
        super().__init__(connection_info)

        self.iopub_socket = self._create_bound_socket('iopub')
        self.shell_socket = self._create_bound_socket('shell')
        self.stdin_socket = self._create_bound_socket('stdin')
        self.control_socket = self._create_bound_socket('control')


class NannyMessaging(MessagingBase):
    socket_types = {
        'nanny_events': zmq.PUB,
        'nanny_control': zmq.ROUTER,
    }

    def __init__(self, connection_info: dict):
        connection_info.setdefault('nanny_events_port', 0)
        connection_info.setdefault('nanny_control_port', 0)
        super().__init__(connection_info)

        self.events_socket = self._create_bound_socket('nanny_events')
        self.control_socket = self._create_bound_socket('nanny_control')



from binascii import b2a_hex
import os
from datetime import timezone, datetime
from ._version import protocol_version

def utcnow():
    """Return timezone-aware UTC timestamp"""
    return datetime.utcnow().replace(tzinfo=timezone.utc)

def new_id():
    """Generate a new random id.

    Avoids problematic runtime import in stdlib uuid on Python 2.

    Returns
    -------

    id string (16 random bytes as hex-encoded text, chunks separated by '-')
    """
    buf = os.urandom(16)
    return u'-'.join(b2a_hex(x).decode('ascii') for x in (
        buf[:4], buf[4:]
    ))

def new_header(msg_type):
    """Create a new message header

    username and session are set later when the message is serialised.
    """
    return {
        'date': utcnow(),
        'msg_type': msg_type,
        'msg_id': new_id(),
        'username': '',
        'session': '',
        'version': protocol_version,
    }

class Message:
    tracker = None  # Set when sending

    def __init__(self, content, header, parent_header=None, metadata=None,
                 idents=None, buffers=None):
        self.content = content
        self.header = header
        self.parent_header = parent_header or {}
        self.metadata = metadata or {}
        self.idents = idents or []
        self.buffers = buffers or []

    @classmethod
    def from_type(cls, msg_type, content, parent_msg=None, metadata=None):
        if parent_msg is None:
            parent_header = {}
            idents = []
        else:
            parent_header = parent_msg.header
            idents = parent_msg.idents
        return cls(content=content, header=new_header(msg_type),
                   parent_header=parent_header, metadata=metadata, idents=idents)

    def make_dict(self):
        return {
            'content': self.content,
            'header': self.header,
            'parent_header': self.parent_header,
            'metadata': self.metadata,
        }

    def __repr__(self):
        args = [repr(self.content), repr(self.header)]
        if self.parent_header:
            args.append('parent_header=%r' % self.parent_header)
        if self.metadata:
            args.append('metadata=%r' % self.parent_header)
        if self.idents:
            args.append('idents=%r' % self.idents)
        if self.buffers:
            args.append('buffers=%r' % self.buffers)
        return "Message(%s)" % ', '.join(args)


def validate_string_dict(dct):
    """Validate that the input is a dict with string keys and values.

    Raises ValueError if not."""
    for k, v in dct.items():
        if not isinstance(k, str):
            raise ValueError('key %r in dict must be a string' % k)
        if not isinstance(v, str):
            raise ValueError('value %r in dict must be a string' % v)

def execute_request(code, silent=False, store_history=True,
            user_expressions=None, allow_stdin=False, stop_on_error=True):
    """Construct an execute_request message.

    Parameters
    ----------
    code : str
        A string of code in the kernel's language.

    silent : bool, optional (default False)
        If set, the kernel will execute the code as quietly possible, and
        will force store_history to be False.

    store_history : bool, optional (default True)
        If set, the kernel will store command history.  This is forced
        to be False if silent is True.

    user_expressions : dict, optional
        A dict mapping names to expressions to be evaluated in the user's
        dict. The expression values are returned as strings formatted using
        :func:`repr`.

    allow_stdin : bool, optional (default self.allow_stdin)
        Flag for whether the kernel can send stdin requests to frontends.

        Some frontends (e.g. the Notebook) do not support stdin requests.
        If raw_input is called from code executed from such a frontend, a
        StdinNotImplementedError will be raised.

    stop_on_error: bool, optional (default True)
        Flag whether to abort the execution queue, if an exception is encountered.
    """
    if user_expressions is None:
        user_expressions = {}

    # Don't waste network traffic if inputs are invalid
    if not isinstance(code, str):
        raise ValueError('code %r must be a string' % code)
    validate_string_dict(user_expressions)

    content = dict(code=code, silent=silent, store_history=store_history,
                   user_expressions=user_expressions,
                   allow_stdin=allow_stdin, stop_on_error=stop_on_error
                   )
    return Message.from_type('execute_request', content)

def shell_reply_error(reply_type, ename, evalue, traceback, *, parent=None):
    """Make a reply to a shell message on an error.

    This has the same structure for different message types. reply_type should
    be e.g. 'execute_reply'.
    """
    return Message.from_type(reply_type, {
        'status': 'error',
        'ename': ename, 'evalue': evalue, 'traceback': traceback,
    }, parent_msg=parent)

def execute_reply(execution_count, payloads=None, user_expressions=None, *, parent=None):
    """Make an execute_reply for successful execution."""
    if payloads is None:
        payloads = []
    if user_expressions is None:
        user_expressions = {}
    return Message.from_type('execute_reply', {
        'status': 'ok',
        'execution_count': execution_count,
        'payloads': payloads,
        'user_expressions': user_expressions,
    }, parent_msg=parent)

def complete_request(code, cursor_pos=None):
    """Make a complete_request message.

    Parameters
    ----------
    code : str
        The context in which completion is requested.
        Can be anything between a variable name and an entire cell.
    cursor_pos : int, optional
        The position of the cursor in the block of code where the completion was requested.
        Default: ``len(code)``
    """
    if cursor_pos is None:
        cursor_pos = len(code)
    content = dict(code=code, cursor_pos=cursor_pos)
    return Message.from_type('complete_request', content)

def complete_reply(matches, cursor_start, cursor_end, metadata=None, *, parent=None):
    """Make a complete_reply for successful completion."""
    if metadata is None:
        metadata = {}
    return Message.from_type('complete_reply', {
        'status': 'ok',
        'matches': matches,
        'cursor_start': cursor_start,
        'cursor_end': cursor_end,
        'metadata': metadata,
    }, parent_msg=parent)

def inspect_request(code, cursor_pos=None, detail_level=0):
    """Make an inspect_request message.

    Get metadata information about an object in the kernel's namespace.

    It is up to the kernel to determine the appropriate object to inspect.

    Parameters
    ----------
    code : str
        The context in which info is requested.
        Can be anything between a variable name and an entire cell.
    cursor_pos : int, optional
        The position of the cursor in the block of code where the info was requested.
        Default: ``len(code)``
    detail_level : int, optional
        The level of detail for the introspection (0-2)
    """
    if cursor_pos is None:
        cursor_pos = len(code)
    return Message.from_type('inspect_request', {
        'code': code,
        'cursor_pos': cursor_pos,
        'detail_level': detail_level,
    })

def inspect_reply(found: bool, data: dict, metadata=None, *, parent=None):
    """Make an inspect_reply for successful introspection."""
    if not found:
        assert not data
    if metadata is None:
        metadata = {}
    return Message.from_type('inspect_reply', {
        'found': found,
        'data': data,
        'metadata': metadata
    }, parent_msg=parent)

def history_request(raw=True, output=False, hist_access_type='range',
                    session=0, start=0, stop=None, n=0, pattern='',
                    unique=False):
    """Get entries from the kernel's history list.

    Parameters
    ----------
    raw : bool
        If True, return the raw input.
    output : bool
        If True, then return the output as well.
    hist_access_type : str
        'range' (fill in session, start and stop params), 'tail' (fill in n)
         or 'search' (fill in pattern param).

    session : int
        For a range request, the session from which to get lines. Session
        numbers are positive integers; negative ones count back from the
        current session.
    start : int
        The first line number of a history range.
    stop : int
        The final (excluded) line number of a history range.

    n : int
        The number of lines of history to get for a tail/search request.

    pattern : str
        The glob-syntax pattern for a search request.
    unique : bool
        Exclude duplicate entries from search results.
    """
    content = {
        'hist_access_type': hist_access_type,
        'raw': raw,
        'output': output,
    }
    if hist_access_type == 'range':
        content.update({
            'session': session,
            'start': start,
            'stop': stop,
        })
    elif hist_access_type == 'tail':
        assert n > 0
        content['n'] = n
    elif hist_access_type == 'search':
        assert pattern != ''
        content.update({
            'pattern': pattern,
            'n': n or None,
            'unique': unique,
        })
    else:
        ValueError("Unknown hist_access_type %r" % hist_access_type)

    return Message.from_type('history_request', content)

def kernel_info_request():
    """Make a kernel_info_request message.
    """
    return Message.from_type('kernel_info_request', {})

def kernel_info_reply(implementation: str, implementation_version: str,
                      language_info: dict, banner='', help_links=None,
                      *, parent=None):
    """Make a kernel_info_reply message"""
    if help_links is None:
        help_links = []
    return Message.from_type('kernel_info_reply', {
        'status': 'ok',
        'protocol_version': protocol_version,
        'implementation': implementation,
        'implementation_version': implementation_version,
        'language_info': language_info,
        'banner': banner,
        'help_links': help_links
    }, parent_msg=parent)

def comm_info_request(target_name=None):
    """Make a comm_info_request message.
    """
    if target_name is None:
        content = {}
    else:
        content = dict(target_name=target_name)
    return Message.from_type('comm_info_request', content)

def comm_info_reply(comms: dict, *, parent=None):
    """Make a comm_info_reply message."""
    return Message.from_type('comm_info_reply', {
        'status': 'ok', 'comms': comms,
    }, parent_msg=parent)

def shutdown_request(restart=False):
    """Make a shutdown_request message.

    Upon receipt of the (empty) reply, client code can safely assume that
    the kernel has shut down and it's safe to forcefully terminate it if
    it's still alive.

    The kernel will send the reply via a function registered with Python's
    atexit module, ensuring it's truly done as the kernel is done with all
    normal operation.
    """
    # Send quit message to kernel. Once we implement kernel-side setattr,
    # this should probably be done that way, but for now this will do.
    return Message.from_type('shutdown_request', {'restart': restart})

def shutdown_reply(restart=False, *, parent=None):
    """Make a shutdown_reply message."""
    return Message.from_type('shutdown_reply', {
        'status': 'ok', 'restart': restart,
    }, parent_msg=parent)

def is_complete_request(code):
    """Make an is_complete_request message.

    This is used to ask the kernel if some code is complete & ready to execute.
    """
    return Message.from_type('is_complete_request', {'code': code})

def is_complete_reply(status, indent=0, *, parent=None):
    """Make an is_complete_reply message."""
    return Message.from_type('is_complete_reply', {
        'status': status, 'indent': indent,
    }, parent_msg=parent)

def interrupt_request():
    """Make an is_complete_request message"""
    return Message.from_type("interrupt_request", content={})

def interrupt_reply(*, parent=None):
    return Message.from_type("interrupt_reply", content={'status': 'ok'},
                             parent_msg=parent)

def input_request(prompt: str, password=False, *, parent=None):
    return Message.from_type("input_request", content={
        'prompt': prompt, 'password': password
    }, parent_msg=parent)

def input_reply(string: str, *, parent=None):
    """Make an input_reply message, to send a string of input to the kernel.

    This should only be called in response to the kernel sending an
    ``input_request`` message on the stdin channel.
    """
    content = dict(value=string)
    return Message.from_type('input_reply', content, parent_msg=parent)

# IOPub messages ---------------

def stream(text, stream_name='stdout', *, parent=None):
    assert stream_name in {'stdout', 'stderr'}
    return Message.from_type('stream', {
        'name': stream_name, 'text': text,
    }, parent_msg=parent)

def display_data(data: dict, metadata: dict =None, transient: dict =None, *, parent=None):
    if metadata is None:
        metadata = {}
    if transient is None:
        transient = {}
    return Message.from_type('display_data', {
        'data': data, 'metadata': metadata, 'transient': transient,
    }, parent_msg=parent)

def update_display_data(data: dict, metadata: dict =None, transient: dict =None, *, parent=None):
    if metadata is None:
        metadata = {}
    if transient is None:
        transient = {}
    return Message.from_type('update_display_data', {
        'data': data, 'metadata': metadata, 'transient': transient,
    }, parent_msg=parent)

def execute_result(execution_count: int, data: dict, metadata: dict =None,
                   transient: dict =None, *, parent=None):
    if metadata is None:
        metadata = {}
    if transient is None:
        transient = {}
    return Message.from_type('execute_result', {
        'execution_count': execution_count,
        'data': data, 'metadata': metadata, 'transient': transient,
    }, parent_msg=parent)

def execute_input(code: str, execution_count: int, *, parent=None):
    return Message.from_type('execute_input', {
        'code': code, 'execution_count': execution_count,
    }, parent_msg=parent)

def error(ename, evalue, traceback, *, parent=None):
    return Message.from_type('error', {
        'ename': ename, 'evalue': evalue, 'traceback': traceback,
    }, parent_msg=parent)

def kernel_status(status: str, *, parent=None):
    assert status in {'busy', 'idle', 'starting'}
    return Message.from_type('status', {
        'execution_state': status,
    }, parent_msg=parent)

def clear_output(wait: bool, *, parent=None):
    return Message.from_type('clear_output',  {'wait': wait}, parent_msg=parent)



"""Session object for serialising, deserialising and signing messages.
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from binascii import b2a_hex
import hashlib
import hmac
import os
import pprint
import random
import warnings

from zmq.utils import jsonapi

from jupyter_protocol.jsonutil import extract_dates, date_default
from ipython_genutils.py3compat import (str_to_bytes, unicode_type,)

from jupyter_protocol.adapter import adapt
from traitlets.log import get_logger
from .messages import Message


#-----------------------------------------------------------------------------
# utility functions
#-----------------------------------------------------------------------------

def squash_unicode(obj):
    """coerce unicode back to bytestrings."""
    if isinstance(obj,dict):
        for key in obj.keys():
            obj[key] = squash_unicode(obj[key])
            if isinstance(key, unicode_type):
                obj[squash_unicode(key)] = obj.pop(key)
    elif isinstance(obj, list):
        for i,v in enumerate(obj):
            obj[i] = squash_unicode(v)
    elif isinstance(obj, unicode_type):
        obj = obj.encode('utf8')
    return obj

#-----------------------------------------------------------------------------
# globals and defaults
#-----------------------------------------------------------------------------

# default values for the thresholds:
MAX_ITEMS = 64
MAX_BYTES = 1024

# ISO8601-ify datetime objects
# allow unicode
# disallow nan, because it's not actually valid JSON
json_packer = lambda obj: jsonapi.dumps(obj, default=date_default,
    ensure_ascii=False, allow_nan=False,
)
json_unpacker = lambda s: jsonapi.loads(s)

DELIM = b"<IDS|MSG>"


def new_id():
    """Generate a new random id.

    Avoids problematic runtime import in stdlib uuid on Python 2.

    Returns
    -------

    id string (16 random bytes as hex-encoded text, chunks separated by '-')
    """
    buf = os.urandom(16)
    return u'-'.join(b2a_hex(x).decode('ascii') for x in (
        buf[:4], buf[4:]
    ))

def new_id_bytes():
    """Return new_id as ascii bytes"""
    return new_id().encode('ascii')

session_aliases = dict(
    ident = 'Session.session',
    user = 'Session.username',
    keyfile = 'Session.keyfile',
)

session_flags  = {
    'secure' : ({'Session' : { 'key' : new_id_bytes(),
                            'keyfile' : '' }},
        """Use HMAC digests for authentication of messages.
        Setting this flag will generate a new UUID to use as the HMAC key.
        """),
    'no-secure' : ({'Session' : { 'key' : b'', 'keyfile' : '' }},
        """Don't authenticate messages."""),
}

def default_secure(cfg):
    """Set the default behavior for a config environment to be secure.

    If Session.key/keyfile have not been set, set Session.key to
    a new random UUID.
    """
    warnings.warn("default_secure is deprecated", DeprecationWarning)
    if 'Session' in cfg:
        if 'key' in cfg.Session or 'keyfile' in cfg.Session:
            return
    # key/keyfile not specified, generate new UUID:
    cfg.Session.key = new_id_bytes()


def extract_header(msg_or_header):
    """Given a message or header, return the header."""
    if not msg_or_header:
        return {}
    try:
        # See if msg_or_header is the entire message.
        h = msg_or_header['header']
    except KeyError:
        try:
            # See if msg_or_header is just the header
            h = msg_or_header['msg_id']
        except KeyError:
            raise
        else:
            h = msg_or_header
    if not isinstance(h, dict):
        h = dict(h)
    return h

class Session(object):
    """Object for handling serialization and sending of messages.

    The Session object handles building messages and sending them
    with ZMQ sockets or ZMQStream objects.  Objects can communicate with each
    other over the network via Session objects, and only need to work with the
    dict-based IPython message spec. The Session will handle
    serialization/deserialization, security, and metadata.

    Sessions support configurable serialization via packer/unpacker traits,
    and signing with HMAC digests via the key/keyfile traits.

    Parameters
    ----------

    key : bytes
        The key used to initialize an HMAC signature.  If unset, messages
        will not be signed or checked.
    signature_scheme : str
        The digest scheme used to construct the message signatures.
        Must have the form 'hmac-HASH' (default: hmac-sha256)
    """
    debug = False   # Turn on for debugging output
    
    # if 0, no adapting to do.
    adapt_version = 0

    # The maximum number of digests to remember.
    # The digest history will be culled when it exceeds this value.
    digest_history_size = 2**16

    def __init__(self, key, signature_scheme='hmac-sha256'):
        self.key = key
        self.signature_scheme = signature_scheme
        self.session_id = new_id()
        self.username = os.environ.get('USER', 'username')
        if self.key:
            hash_name = self.signature_scheme.split('-', 1)[1]
            digest_mod = getattr(hashlib, hash_name)
            self.auth = hmac.HMAC(self.key, digestmod=digest_mod)
        else:
            self.auth = None
            get_logger().warning("Message signing is disabled.  This is insecure and not recommended!")
        self.digest_history = set()

    def clone(self):
        """Create a copy of this Session

        Useful when connecting multiple times to a given kernel.
        This prevents a shared digest_history warning about duplicate digests
        due to multiple connections to IOPub in the same process.
        """
        # make a copy
        new_session = type(self)(key=self.key)
        # fork digest_history
        new_session.digest_history.update(self.digest_history)
        return new_session

    def sign(self, msg_list):
        """Sign a message with HMAC digest. If no auth, return b''.

        Parameters
        ----------
        msg_list : list
            The [p_header,p_parent,p_content] part of the message list.
        """
        if self.auth is None:
            return b''
        h = self.auth.copy()
        for m in msg_list:
            h.update(m)
        return str_to_bytes(h.hexdigest())

    def serialize(self, msg):
        """Serialize the message components to bytes.

        This is roughly the inverse of deserialize. The serialize/deserialize
        methods work with full message lists, whereas pack/unpack work with
        the individual message parts in the message list.

        Parameters
        ----------
        msg : Message
            The message object to be sent

        Returns
        -------
        msg_list : list
            The list of bytes objects to be sent with the format::

                [ident1, ident2, ..., DELIM, HMAC, p_header, p_parent,
                 p_metadata, p_content, buffer1, buffer2, ...]

            In this list, the ``p_*`` entities are the packed or serialized
            versions, so if JSON is used, these are utf8 encoded JSON strings.
        """
        msg_dict = msg.make_dict()

        if self.adapt_version:
            msg_dict = adapt(msg_dict, self.adapt_version)

        header = msg_dict['header'] #.copy()
        header['session'] = self.session_id
        header['username'] = self.username

        real_message = [json_packer(header),
                        json_packer(msg_dict['parent_header']),
                        json_packer(msg_dict['metadata']),
                        json_packer(msg_dict['content']),
        ]

        to_send = msg.idents + [DELIM, self.sign(real_message)] + real_message \
                    + msg.buffers

        return to_send

    def feed_identities(self, msg_list, copy=True):
        """Split the identities from the rest of the message.

        Feed until DELIM is reached, then return the prefix as idents and
        remainder as msg_list. This is easily broken by setting an IDENT to DELIM,
        but that would be silly.

        Parameters
        ----------
        msg_list : a list of zmq.Message or bytes objects
            The message to be split.
        copy : bool
            flag determining whether the arguments are bytes or Messages

        Returns
        -------
        (idents, msg_list) : two lists
            idents will always be a list of bytes, each of which is a ZMQ
            identity. msg_list will be a list of bytes or zmq.Messages of the
            form [HMAC,p_header,p_parent,p_content,buffer1,buffer2,...] and
            should be unpackable/unserializable via self.deserialize at this
            point.
        """
        if copy:
            idx = msg_list.index(DELIM)
            return msg_list[:idx], msg_list[idx+1:]
        else:
            failed = True
            for idx,m in enumerate(msg_list):
                if m.bytes == DELIM:
                    failed = False
                    break
            if failed:
                raise ValueError("DELIM not in msg_list")
            idents, msg_list = msg_list[:idx], msg_list[idx+1:]
            return [m.bytes for m in idents], msg_list

    def _add_digest(self, signature):
        """add a digest to history to protect against replay attacks"""
        if self.digest_history_size == 0:
            # no history, never add digests
            return

        self.digest_history.add(signature)
        if len(self.digest_history) > self.digest_history_size:
            # threshold reached, cull 10%
            self._cull_digest_history()

    def _cull_digest_history(self):
        """cull the digest history

        Removes a randomly selected 10% of the digest history
        """
        current = len(self.digest_history)
        n_to_cull = max(int(current // 10), current - self.digest_history_size)
        if n_to_cull >= current:
            self.digest_history = set()
            return
        to_cull = random.sample(self.digest_history, n_to_cull)
        self.digest_history.difference_update(to_cull)

    def deserialize_msg_parts(self, msg_list, content=True, copy=True):
        """Unserialize a msg_list to a nested message dict.

        This is roughly the inverse of serialize. The serialize/deserialize
        methods work with full message lists, whereas pack/unpack work with
        the individual message parts in the message list.

        Parameters
        ----------
        msg_list : list of bytes or zmq.Message objects
            The list of message parts of the form [HMAC,p_header,p_parent,
            p_metadata,p_content,buffer1,buffer2,...].
        content : bool (True)
            Whether to unpack the content dict (True), or leave it packed
            (False).
        copy : bool (True)
            Whether msg_list contains bytes (True) or the non-copying Message
            objects in each place (False).

        Returns
        -------
        msg : dict
            The nested message dict with top-level keys [header, parent_header,
            content, buffers].  The buffers are returned as memoryviews.
        """
        minlen = 5
        message = {}
        if not copy:
            # pyzmq didn't copy the first parts of the message, so we'll do it
            for i in range(minlen):
                msg_list[i] = msg_list[i].bytes
        if self.auth is not None:
            signature = msg_list[0]
            if not signature:
                raise ValueError("Unsigned Message")
            if signature in self.digest_history:
                raise ValueError("Duplicate Signature: %r" % signature)
            if content:
                # Only store signature if we are unpacking content, don't store if just peeking.
                self._add_digest(signature)
            check = self.sign(msg_list[1:5])
            if not hmac.compare_digest(signature, check):
                raise ValueError("Invalid Signature: %r" % signature)
        if not len(msg_list) >= minlen:
            raise TypeError("malformed message, must have at least %i elements"%minlen)
        header = json_unpacker(msg_list[1])
        message['header'] = extract_dates(header)
        message['parent_header'] = extract_dates(json_unpacker(msg_list[2]))
        message['metadata'] = json_unpacker(msg_list[3])
        if content:
            message['content'] = json_unpacker(msg_list[4])
        else:
            message['content'] = msg_list[4]
        buffers = [memoryview(b) for b in msg_list[5:]]
        if buffers and buffers[0].shape is None:
            # force copy to workaround pyzmq #646
            buffers = [memoryview(b.bytes) for b in msg_list[5:]]
        message['buffers'] = buffers
        if self.debug:
            pprint.pprint(message)
        # adapt to the current version
        return adapt(message)

    def deserialize(self, raw_msg, content=True, copy=True):
        """Deserialize a message from the full wire format.

        This combines feed_identities and deserialize_msg_parts()
        """
        idents, msg_list = self.feed_identities(raw_msg, copy)
        msg_dict = self.deserialize_msg_parts(msg_list, content=content, copy=copy)
        msg_dict['idents'] = idents
        return Message(**msg_dict)



"""test building messages with Session"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import hmac
import os
import sys
import uuid
from datetime import datetime
try:
    from unittest import mock
except ImportError:
    import mock

import pytest

import zmq

from zmq.tests import BaseZMQTestCase

from jupyter_protocol import session as ss
from jupyter_protocol import jsonutil
from jupyter_protocol.messages import utcnow, Message, new_header

from ipython_genutils.py3compat import string_types

def _bad_packer(obj):
    raise TypeError("I don't work")

def _bad_unpacker(bytes):
    raise TypeError("I don't work either")

class SessionTestCase(BaseZMQTestCase):

    def setUp(self):
        BaseZMQTestCase.setUp(self)
        self.session = ss.Session(key=b'')


@pytest.fixture
def no_copy_threshold():
    """Disable zero-copy optimizations in pyzmq >= 17"""
    with mock.patch.object(zmq, 'COPY_THRESHOLD', 1):
        yield


@pytest.mark.usefixtures('no_copy_threshold')
class TestSession(SessionTestCase):

    def test_serialize(self):
        msg = Message.from_type('execute', content=dict(a=10, b=1.1))
        msg.idents = [b'foo']
        msg_list = self.session.serialize(msg)
        new_msg = self.session.deserialize(msg_list)
        assert new_msg.idents == [b'foo']
        assert new_msg.header == msg.header
        assert new_msg.content == msg.content
        assert new_msg.parent_header == msg.parent_header
        assert new_msg.metadata == msg.metadata
        # ensure floats don't come out as Decimal:
        assert type(new_msg.content['b']) == type(new_msg.content['b'])

    def test_unique_msg_ids(self):
        """test that messages receive unique ids"""
        ids = set()
        for i in range(2**12):
            h = new_header('test')
            msg_id = h['msg_id']
            self.assertTrue(msg_id not in ids)
            ids.add(msg_id)

    def test_zero_digest_history(self):
        session = ss.Session(key=b'')
        session.digest_history_size = 0
        for i in range(11):
            session._add_digest(uuid.uuid4().bytes)
        self.assertEqual(len(session.digest_history), 0)

    def test_cull_digest_history(self):
        session = ss.Session(key=b'')
        session.digest_history_size = 100
        for i in range(100):
            session._add_digest(uuid.uuid4().bytes)
        assert len(session.digest_history) == 100
        session._add_digest(uuid.uuid4().bytes)
        assert len(session.digest_history) == 91
        for i in range(9):
            session._add_digest(uuid.uuid4().bytes)
        assert len(session.digest_history) == 100
        session._add_digest(uuid.uuid4().bytes)
        assert len(session.digest_history) == 91

    def test_datetimes(self):
        content = dict(t=utcnow())
        metadata = dict(t=utcnow())
        p = Message.from_type('msg', {})
        msg = Message.from_type('msg', content, metadata=metadata, parent_msg=p)
        smsg = self.session.serialize(msg)
        msg2 = self.session.deserialize(smsg)
        assert isinstance(msg2.header['date'], datetime)
        assert msg.header == msg2.header
        assert msg.parent_header == msg2.parent_header
        assert msg.parent_header == msg2.parent_header
        assert isinstance(msg.content['t'], datetime)
        assert isinstance(msg.metadata['t'], datetime)
        assert isinstance(msg2.content['t'], string_types)
        assert isinstance(msg2.metadata['t'], string_types)
        assert msg.content == jsonutil.extract_dates(msg2.content)
        assert msg.metadata == jsonutil.extract_dates(msg2.metadata)
    
    def test_clone(self):
        s = self.session
        s._add_digest('initial')
        s2 = s.clone()
        assert s2.digest_history == s.digest_history
        assert s2.digest_history is not s.digest_history
        digest = 'abcdef'
        s._add_digest(digest)
        assert digest in s.digest_history
        assert digest not in s2.digest_history



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function

# the name of the project
name = 'jupyter_protocol'

import os
import sys
from setuptools import setup
from setuptools.command.bdist_egg import bdist_egg

v = sys.version_info
if v[:2] < (3, 5):
    error = "ERROR: %s requires Python version 3.5 or above." % name
    print(error, file=sys.stderr)
    sys.exit(1)

pjoin = os.path.join
here = os.path.abspath(os.path.dirname(__file__))
pkg_root = pjoin(here, name)

packages = []
for d, _, _ in os.walk(pjoin(here, name)):
    if os.path.exists(pjoin(d, '__init__.py')):
        packages.append(d[len(here)+1:].replace(os.path.sep, '.'))

version_ns = {}
with open(pjoin(here, name, '_version.py')) as f:
    exec(f.read(), {}, version_ns)

class bdist_egg_disabled(bdist_egg):
    """Disabled version of bdist_egg

    Prevents setup.py install from performing setuptools' default easy_install,
    which it should never ever do.
    """
    def run(self):
        sys.exit("Aborting implicit building of eggs. Use `pip install .` to install from source.")


setup_args = dict(
    name            = name,
    version         = version_ns['__version__'],
    packages        = packages,
    description     = "Jupyter protocol implementation",
    author          = 'Jupyter Development Team',
    author_email    = 'jupyter@googlegroups.com',
    url             = 'https://jupyter.org',
    license         = 'BSD',
    classifiers     = [
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    install_requires = [
        'traitlets',
        'jupyter_core',
        'pyzmq>=13',
        'python-dateutil>=2.1',
    ],
    extras_require   = {
        'test': ['ipykernel', 'ipython', 'mock', 'pytest'],
    },
    python_requires = ">=3.5",
    cmdclass         = {
        'bdist_egg': bdist_egg if 'bdist_egg' in sys.argv else bdist_egg_disabled,
    },
)


if __name__ == '__main__':
    setup(**setup_args)



import os
import pytest
import zmq
from zmq.eventloop.zmqstream import ZMQStream
from zmq.tests import BaseZMQTestCase

from jupyter_protocol.messages import Message
from jupyter_protocol.session import Session
from jupyter_protocol.sockets import MessagingBase, DONE

def _messaging_w_socket_pair(a, b):
    messaging = MessagingBase({
        'key': '',
        'signature_scheme': 'hmac-sha256',
    })
    messaging.sockets['A'] = a
    messaging.sockets['B'] = b
    return messaging

def test_send():
    ctx = zmq.Context.instance()
    A = ctx.socket(zmq.PAIR)
    B = ctx.socket(zmq.PAIR)
    A.bind("inproc://test")
    B.connect("inproc://test")

    messaging = _messaging_w_socket_pair(A, B)

    msg = Message.from_type('test_xmpl', dict(a=10))
    msg.idents = [b'foo']
    msg.buffers = [b'bar']

    messaging.send('A', msg)

    new_msg = messaging.recv('B')
    assert new_msg.idents[0] == b'foo'
    assert new_msg.header == msg.header
    assert new_msg.content == msg.content
    assert new_msg.parent_header == msg.parent_header
    assert new_msg.metadata == msg.metadata
    assert new_msg.buffers == msg.buffers

    # buffers must support the buffer protocol
    msg.buffers = [1]
    with pytest.raises(TypeError):
        messaging.send('A', msg)

    # buffers must be contiguous
    buf = memoryview(os.urandom(16))
    msg.buffers = [buf[::2]]
    with pytest.raises(ValueError):
        messaging.send('A', msg)

    A.close()
    B.close()
    ctx.term()

# def test_send_raw():
#     ctx = zmq.Context.instance()
#     A = ctx.socket(zmq.PAIR)
#     B = ctx.socket(zmq.PAIR)
#     A.bind("inproc://test")
#     B.connect("inproc://test")
#
#     msg = self.session.msg('execute', content=dict(a=10))
#     msg_list = [self.session.pack(msg[part]) for part in
#                 ['header', 'parent_header', 'metadata', 'content']]
#     self.session.send_raw(A, msg_list, ident=b'foo')
#
#     ident, new_msg_list = self.session.feed_identities(B.recv_multipart())
#     new_msg = self.session.deserialize(new_msg_list)
#     self.assertEqual(ident[0], b'foo')
#     self.assertEqual(new_msg['msg_type'],msg['msg_type'])
#     self.assertEqual(new_msg['header'],msg['header'])
#     self.assertEqual(new_msg['parent_header'],msg['parent_header'])
#     self.assertEqual(new_msg['content'],msg['content'])
#     self.assertEqual(new_msg['metadata'],msg['metadata'])
#
#     A.close()
#     B.close()
#     ctx.term()

class SocketsTestCase(BaseZMQTestCase):
    def test_tracking(self):
        """test tracking messages"""
        a,b = self.create_bound_pair(zmq.PAIR, zmq.PAIR)
        messaging = _messaging_w_socket_pair(a, b)
        messaging.copy_threshold = 1
        stream = ZMQStream(a)
        msg = Message.from_type('test_xmpl', {})
        messaging.send('A', msg, track=False)
        assert msg.tracker is DONE

        messaging.send('A', msg, track=True)
        assert isinstance(msg.tracker, zmq.MessageTracker)

        msg2 = Message.from_type('test_xmpl', {})
        msg2.buffers = [zmq.Frame(b'hi there', track=True, copy=False)]
        messaging.send('A', msg2, track=True)
        t = msg2.tracker
        assert isinstance(t, zmq.MessageTracker)
        assert t is not DONE
        with pytest.raises(zmq.NotDone):
            t.wait(.1)
        del msg2
        t.wait(1)
